"""
    The 'inspect' module (native in python) is very interesting because
    it has some methods to check if a user defined function or things 
    like that.

    'f.__qualname__' is very interesting thing as it works
    with the full path, not as 'f.__name__' does, this is
    one example of it value when doing it from a local file:
    
    'Example.example.__qualname__'
    test_discord_video.<locals>.Example.example
    
    And yes, the 'example' is a @staticmethod defined in the
    Example class that is contained in the 'test_discord_video'
    file
"""
from yta_general_utils.programming.error_message import ErrorMessage
# from yta_general_utils.file.filename import filename_is_type
# from yta_general_utils.checker.type import variable_is_positive_number
# from yta_general_utils.file.checker import file_exists
# from yta_general_utils.checker.url import url_is_ok
from enum import Enum
from typing import Union

import numpy as np
import inspect
import re


# TODO: Remove commented methods in PythonValidator for the next version
class PythonValidator:
    """
    Class to simplify and encapsulate the functionality related with
    parameters and variables validation.

    This class has been created to simplify the way it work and 
    replace the old ParameterValidator that was using too many 
    different methods being imported and generating cyclic import
    issues.

    We have some equivalent methods that do not need to pass the class
    as a parameter, so we can only work with the class name and avoid
    imports that can cause issues.
    """

    @staticmethod
    def is_instance(element, cls: Union[str, type, list]):
        """
        Check if the provided 'element' is an instance of the provided
        class (or classes) 'cls'. An instance is not the same as a
        class. The 'cls' parameter can be the class or the string name
        of that class, or a list of them (even mixed).

        This method is useful if you want to check if it belongs to a
        class but without importing the class and passing it as a 
        parameter to avoid cyclic import issues.
        """
        if not PythonValidator.is_list(cls):
            cls = [cls]

        # TODO: I need to review this below when providing Unions or
        # strange things because of the 'validate_method_params'
        # method
        if any(not PythonValidator.is_string(cls_item) and not PythonValidator.is_a_class(cls_item) and cls_item is not None for cls_item in cls):
            classes_str = [cls_item.__str__() for cls_item in cls]
            #classes_str = ', '.join(cls)
            raise Exception(f'At least one of the provided "cls" parameters "{classes_str}" is not a string nor a class.')
        
        if PythonValidator.is_a_class(element):
            return False
        
        cls_class_items = tuple([cls_item for cls_item in cls if PythonValidator.is_a_class(cls_item)])
        cls_str_items = [cls_item for cls_item in cls if not PythonValidator.is_a_class(cls_item)]

        return isinstance(element, cls_class_items) or any(getattr(type(element), '__name__', None) == cls_item for cls_item in cls_str_items)
    
    @staticmethod
    def is_an_instance(element):
        """
        Check if the provided 'element' is an instance of any class.
        """
        return isinstance(element, object)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is None
    
    @staticmethod
    def is_class(element, cls: Union[str, type]):
        """
        Check if the provided 'element' is the provided class 'cls'.
        A class is not the same as an instance of that class. The
        'cls' parameter can be the class or the string name of that
        class. 

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if not PythonValidator.is_string(cls) and not PythonValidator.is_a_class(cls):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if PythonValidator.is_string(cls):
            if cls == 'str':
                return PythonValidator.is_string(element)

            # TODO: This is not working at all, I need to check why
            # and to apply more ifs (maybe)
            return getattr(element, '__name__', None) is cls
        else:
            return PythonValidator.is_a_class(element) and element == cls

    @staticmethod
    def is_a_class(element):
        """
        Check if the provided 'element' is a class.
        """
        return isinstance(element, type)
        # This below is just another alternative I want to keep
        return getattr(element, '__name__', None) is not None
    
    @staticmethod
    def is_subclass(element: type, cls: Union[str, type]):
        """
        Check if the provided 'element' is a subclass of the provided
        class 'cls'. The 'cls' parameter can be the class or the
        string name of that class.

        You can pass the string name of the class to avoid import in 
        the file where you are calling this method.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: Convert this to accept cls as array and check 
        # diferent classes
        if not PythonValidator.is_string(cls) and not PythonValidator.is_a_class(cls):
            raise Exception(f'The provided "cls" parameter "{str(cls)}" is not a string nor a class.')
        
        if not PythonValidator.is_a_class(element):
            # We want to know if it is a subclass, we don't
            # care if the provided one is an instance or a class,
            # so we try with both.
            if PythonValidator.is_an_instance(element):
                element = type(element)
            else:
                return False
            
        """
        The element.__mro__ returns a list like this below in
        which you can see, oredered, the classes hierarchy.
        This can be useful, so I keep it here:

        (<class 'yta_multimedia.video.generation.manim.classes.text.simple_text_manim_animation.ExampleManimAnimationGenerator'>, <class 'yta_multimedia.video.generation.manim.classes.base_manim_animation.BaseManimAnimation'>, <class 'manim.scene.scene.Scene'>, <class 'object'>)    
        """
            
        if PythonValidator.is_string(cls):
            return cls in [base_class.__name__ for base_class in element.__bases__]
        else: 
            return issubclass(element, cls)

    @staticmethod
    def is_a_function(element):
        """
        Check if the provided 'element' is a function.
        """
        # TODO: Maybe inspect.isfunction(element) (?)
        return type(element).__name__ == 'function'
    
    @staticmethod
    def is_class_staticmethod(cls: type, method: 'function', method_name: str = None):
        """
        Check if the provided 'method' is an staticmethod (a
        function) defined in the also provided 'cls' class.

        If the 'method_name' parameter is provided, it will
        also check if the name of the provided 'method' is
        equal to the one provided in the 'method_name' param.
        """
        if not PythonValidator.is_a_class(cls):
            raise Exception('The provided "cls" parameter is not a class.')
        
        if not PythonValidator.is_a_function(method):
            raise Exception('The provided "method" parameter is not a function.')
        
        for function in inspect.getmembers(cls, predicate = inspect.isfunction):
            if function[0] == method.__name__:
                if method_name is None:
                    return True
                else:
                    return method.__name__ == method_name
            
        return False
    
    @staticmethod
    def is_list(element):
        """
        Check if the provided 'element' is a list, which
        is quite different from an array.
        """
        return type(element) == list
    
    @staticmethod
    def is_array(element):
        """
        Check if the provided 'element' is an array, which
        is quite different from a list.
        """
        from array import array

        return type(element) == array

    @staticmethod
    def is_empty_list(element):
        """
        Check if the provided 'element' is a list but empty.
        """        
        return PythonValidator.is_list(element) and len(element) == 0
    
    @staticmethod
    def is_list_of(element, cls: type):
        """
        Check if the provided 'element' is a list in which
        all the items are instances of the provided 'cls'
        class.
        """
        return PythonValidator.is_list(element) and all(PythonValidator.is_instance(item, cls) for item in element)
    
    @staticmethod
    def is_list_of_string(element):
        """
        Check if the provided 'element' is a list in which
        all the items are strings.
        """
        return PythonValidator.is_list(element) and all(PythonValidator.is_string(item) for item in element)

    @staticmethod
    def is_dict(element):
        """
        Check if the provided 'element' is a dict.
        """
        return PythonValidator.is_instance(element, dict)

    @staticmethod
    def is_tuple(element):
        """
        Check if the provided 'element' is a tuple.
        """
        return PythonValidator.is_instance(element, tuple)
    
    @staticmethod
    def is_tuple_or_list_or_array_of_n_elements(element, n: int):
        """
        Check if the provided 'element' is a tuple or a list
        with 'n' values.
        """
        return (
            (
                PythonValidator.is_tuple(element) or 
                PythonValidator.is_list(element) or
                PythonValidator.is_array(element)
            ) and
            len(element) == n
        )
    
    @staticmethod
    def is_string(element):
        """
        Check if the provided 'element' is a string (str).
        """
        return isinstance(element, str)
    
    @staticmethod
    def is_boolean(element):
        """
        Check if the provided 'element' is a boolean (bool).
        """
        return isinstance(element, bool)
    
    @staticmethod
    def is_number(element, do_accept_string_number: bool = False):
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        return NumberValidator.is_number(element, do_accept_string_number)
    
    @staticmethod
    def is_numpy_array(element):
        """
        Check if the provided 'element' is an instance of the
        numpy array np.ndarray.
        """
        return PythonValidator.is_instance(element, np.ndarray)
    
    @staticmethod
    def is_enum(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a subclass of an Enum or
        a YTAEnum.

        This method can return True if the provided 'element' is an
        instance or a class that inherits from Enum or YTAEnum class.
        Consider this above to check later if your 'element' is an
        instance or a class.
        """
        # TODO: I think it is 'EnumMeta' not Enum
        return PythonValidator.is_subclass(element, 'YTAEnum') or PythonValidator.is_subclass(element, 'Enum')
    
    @staticmethod
    def is_enum_instance(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is an instance.
        """
        return PythonValidator.is_enum(element) and PythonValidator.is_an_instance(element)
    
    @staticmethod
    def is_enum_class(element: Union['YTAEnum', Enum]):
        """
        Check if the provided 'element' is a Enum (it is a subclass
        of an Enum or a YTAEnum) and it is a class.
        """
        return PythonValidator.is_enum(element) and PythonValidator.is_a_class(element)
    
    # TODO: This method below is not complete, need work
    @staticmethod
    def validate_method_params(method: 'function', params: list, params_to_ignore: list[str] = ['self', 'cls', 'args', 'kwargs']):
        """
        IMPORTANT! This method should be called on top of any method
        in which you want to validate if the provided parameters are
        valid, by providing the method (function) declaration and
        also the 'locals()' function executed as 'params' parameter.
        So, it should be called like this:

        PythonValidator.validate_method_params(function, locals())

        This method check the types of the params that the provided
        'method' has and validate if the provided values fit the
        specified types (according also to the default values). It
        will raise an Exception when any of the provided params (and
        not ignored) is not valid according to its type declaration.

        The provided 'params' must be a dict containing all the param
        names and values.

        This method is able to parse non-type declarations, as in
        "method(type)", single declarations, as in "method(type: str)"
        and Union declarations, as in
        "method(type: Union[str, MyClass])"

        The 'method' parameter must be a real python method to be able
        to inspect it.
        """
        if not PythonValidator.is_a_function(method):
            raise Exception('The provided "method" parameter is not a valid method (function).')

        # This below is like '<class 'package.of.class.name'> or '<class 'str'>
        SINGLE_TYPE_REGEX = r"<class '([^']+)'>"
        # This below is like 'Union['str', 'int', 'FfmpegHandler']
        UNION_TYPE_REGEX = r"typing\.Union\[\s*((?:[^,]+(?:\s*,\s*)?)+)\s*\]"
        # This below is to flag those params with no default value
        # because None can be a default value indicating that it is
        # an optional value
        NO_DEFAULT_VALUE = '__no_default_value__'
        NO_TYPE = '__no_type__'

        # TODO: Refactor this below to make it easier to be read
        for param in inspect.signature(method).parameters.values():
            if param.name in params_to_ignore:
                continue

            print(param.name)
            print(params_to_ignore)

            types = param.annotation if param.annotation is not inspect.Parameter.empty else NO_TYPE
            default_value = param.default if param.default is not inspect.Parameter.empty else NO_DEFAULT_VALUE

            # 'types' can be nothing, a single type or an Union
            if types:
                match_class = re.match(SINGLE_TYPE_REGEX, str(types))
                match_union = re.match(UNION_TYPE_REGEX, str(types))

                # Turn type to array of string types
                if match_class:
                    types = [match_class.group(1).split('.')[-1]]
                elif match_union:
                    classes = match_union.group(1).split(',')
                    types = [class_i.strip().split('.')[-1] for class_i in classes]

            # Now check with the param provided
            user_param = params.get(param.name, None)

            if types is NO_TYPE and user_param is None:
                # If no type we cannot validate anything, but I 
                # think if no type nor value it will not be
                # executed and this Exception below will never
                # happen
                raise Exception(f'The param "{str(param.name)}" has no type declaration but also None value, so it is not accepted.')
            elif user_param == None and (PythonValidator.is_list(types) and 'None' in types or default_value == None):
                # TODO: If we are strict with typing, a value that
                # can be None should be Union[None, ...] and also
                # param = None (as default value) to indicate it is
                # optional, but we accept both of them separately

                # Param value is None and None is accepted or is
                # default value
                pass
            elif user_param == None:
                raise Exception(f'The param "{str(param.name)}" has None value provided and we expected one of these types: {", ".join(types)}.')
            else:
                if not PythonValidator.is_instance(user_param, types) and types is not NO_TYPE:
                    print(types)
                    types_str = ', '.join(types)
                    raise Exception(f'The param value "{str(param.name)}" provided "{str(user_param)}" is not one of the expected types: {types_str}.')

        return True

class NumberValidator:
    """
    Class to simplify and encapsulate the functionality related
    to validate numeric values.
    """
    @staticmethod
    def is_number(element: Union[int, float, str], do_accept_string_number: bool = False):
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        if not PythonValidator.is_instance(element, [int, float, str, np.number]):
            return False
        
        if PythonValidator.is_instance(element, str):
            if do_accept_string_number:
                try:
                    float(element)
                except:
                    return False
            else:
                return False
            
        return True
    
    @staticmethod
    def is_positive_number(element: Union[int, float, str], do_include_zero: bool = True):
        """
        This method checks if the provided 'element' is a numeric type,
        or tries to cast it as a float number if string provided, and
        returns True in the only case that the 'element' is actual a
        number by itself or as a string and it is 0 or above it. If 
        'do_include_zero' is set to False it won't be included.
        """        
        if not NumberValidator.is_number(element, False):
            return False
        
        element = float(element)

        if do_include_zero:
            return element >= 0
        
        return element > 0
    
    @staticmethod
    def is_number_between(element: Union[int, float, str], lower_limit: Union[int, float, str], upper_limit: Union[int, float, str], do_include_lower_limit: bool = True, do_include_upper_limit: bool = True):
        """
        This methods returns True if the provided 'variable' is a valid number
        that is between the also provided 'lower_limit' and 'upper_limit'. It
        will return False in any other case.
        """
        if not NumberValidator.is_number(element, True):
            return False
        
        if not NumberValidator.is_number(lower_limit) or not NumberValidator.is_number(upper_limit):
            return False
        
        element = float(element)
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)
        
        # TODO: Should we switch limits if unordered (?)
        # if upper_limit < lower_limit:
        #     raise Exception(f'The provided "upper_limit" parameter {str(upper_limit)} is lower than the "lower_limit" parameter {str(lower_limit)} provided.')

        if do_include_lower_limit and do_include_upper_limit:
            return lower_limit <= element <= upper_limit
        elif do_include_lower_limit:
            return lower_limit <= element < upper_limit
        elif do_include_upper_limit:
            return lower_limit < element <= upper_limit
        else:
            return lower_limit < element < upper_limit
        
    @staticmethod
    def is_int(element: int):
        """
        Return True if the provided 'element' is an int
        number.
        """
        return PythonValidator.is_instance(element, int)
    
    @staticmethod
    def is_float(element: float):
        """
        Return True if the provided 'element' is a float
        number.
        """
        return PythonValidator.is_instance(element, float)

    @staticmethod
    def is_even(element: float):
        """
        Return True if the provided 'element' is an even
        number. This method considers that the provided
        'element' is a valid number.
        """
        return element % 2 == 0
    
    @staticmethod
    def is_odd(element: float):
        """
        Return True if the provided 'element' is an odd
        number. This method considers that the provided
        'element' is a valid number.
        """
        return element % 2 != 0
    
# TODO: This 'ParameterValidator' is no longer used because
# of the new PythonValidator, but it could be useful for
# simple validations like 'mandatory', etc. including the
# error message. Complex methods have been commented due
# to cyclic import issues, so avoiding them we could have
# an easy way to validate parameters.
class ParameterValidator:
    @staticmethod
    def validate_mandatory_parameter(name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' has a value, raising an Exception if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not value:
            raise Exception(ErrorMessage.parameter_not_provided(name))

        return value

    @staticmethod
    def validate_string_parameter(name: str, value):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a string value, raising an Exception if
        not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not PythonValidator.is_string(value):
            raise Exception(ErrorMessage.parameter_is_not_string(name))

        return value
    
    @staticmethod
    def validate_bool_parameter(name: str, value: bool):
        """
        Validates if the provided 'value' parameter with the also
        provided 'name' is a boolean value, raising and Exception 
        if not.

        This method returns the provided 'value' if everything is
        ok.
        """
        if not PythonValidator.is_boolean(value):
            raise Exception(ErrorMessage.parameter_is_not_boolean(name))

        return value
    
# class ParameterValidator:
#     @classmethod
#     def validate_mandatory_parameter(cls, name: str, value):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' has a value, raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not value:
#             raise Exception(ErrorMessage.parameter_not_provided(name))

#         return value
        
#     @classmethod
#     def validate_string_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a string value, raising an Exception if
#         not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not isinstance(value, str):
#             raise Exception(ErrorMessage.parameter_is_not_string(name))

#         return value

#     @classmethod
#     def validate_bool_parameter(cls, name: str, value: bool):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a boolean value, raising and Exception 
#         if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not isinstance(value, bool):
#             raise Exception(ErrorMessage.parameter_is_not_boolean(name))

#         return value
        
#     @classmethod
#     def validate_file_exists(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a file that actually exists, raising
#         an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not file_exists(value):
#             raise Exception(ErrorMessage.parameter_is_file_that_doesnt_exist(name))

#         return value
        
#     @classmethod
#     def validate_filename_is_type(cls, name: str, value: str, file_type: 'FileType'):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a filename of the given 'file_type',
#         raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not filename_is_type(value, file_type):
#             raise Exception(ErrorMessage.parameter_is_not_file_of_file_type(name, file_type))

#         return value
        
#     @classmethod
#     def validate_url_is_ok(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a valid url (the url is accessible),
#         raising an Exception if not.

#         This method returns the provided 'value' if everything is
#         ok.
#         """
#         if not url_is_ok(value):
#             raise Exception(ErrorMessage.parameter_is_not_valid_url(name))

#         return value
        
#     @classmethod
#     def validate_positive_number(cls, name: str, value: Union[int, float]):
#         """
#         Validates if the provided 'value' parameter with the also
#         provided 'name' is a positive number (0 or greater),
#         raising an Exception if not.

#         This method returns the provided 'value' as it is if 
#         everything is ok.
#         """
#         if not variable_is_positive_number(value):
#             raise Exception(ErrorMessage.parameter_is_not_positive_number(name))

#         return value

#     @classmethod
#     def validate_is_class(cls, name: str, value, class_names: Union[list[str], str]):
#         """
#         Validates if the provided 'value' is one of the provided 'class_names'
#         by using the 'type(value).__name__' checking.

#         This method returns the 'value' as it is if everything is ok.
#         """
#         if isinstance(class_names, str):
#             class_names = [class_names]

#         if not type(value).__name__ in class_names:
#             raise Exception(ErrorMessage.parameter_is_not_class(name, class_names))
        
#         return value
        
#     # Complex ones below
#     @classmethod
#     def validate_string_mandatory_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' is a valid and non
#         empty string.
#         """
#         cls.validate_mandatory_parameter(name, value)
#         cls.validate_string_parameter(name, value)

#         return value

#     @classmethod
#     def validate_numeric_positive_mandatory_parameter(cls, name: str, value: str):
#         """
#         Validates if the provided 'value' is a positive numeric
#         value.
#         """
#         cls.validate_mandatory_parameter(name, value)
#         cls.validate_positive_number(name, value)

#         return value
    
#     @classmethod
#     def validate_is_enum_class(cls, enum: Union['YTAEnum', Enum]):
#         """
#         Validates if the provided 'value' is a valid Enum
#         class or subclass.

#         This method will raise an Exception if the provided
#         'value' is not a valid Enum class or subclass, or
#         will return it as it is if yes.
#         """
#         if not isinstance(enum, Enum) and not issubclass(enum, Enum):
#             raise Exception(f'The parameter "{enum}" provided is not an Enum class or subclass.')
        
#         return enum

#     @classmethod
#     def validate_enum(cls, value: Union['YTAEnum', str], enum: 'YTAEnum'):
#         """
#         Validates if the provided 'value' value is a valid
#         Enum or Enum value of the also provided 'enum' class.

#         This method will raise an Exception if something is
#         wrong or will return the 'value' as an 'enum' Enum.
#         instance if everything is ok.
#         """
#         cls.validate_mandatory_parameter('value', value)
#         cls.validate_is_enum_class(enum)
#         cls.validate_is_class('value', value, [enum.__class__.__name__, 'str'])

#         return enum.to_enum(value)


from requests import Response

import json


class FileWriter:
    """
    Class to simplify and encapsulate the functionality related with
    writting files.
    """
    @staticmethod
    def write_binary_file(binary_data: bytes, filename: str):
        """
        Writes the provided 'binary_data' in the 'filename' file. It replaces the
        previous content if existing.
        """
        if not binary_data:
            return None
        
        if not filename:
            return None
        
        f = open(filename, 'wb')
        f.write(binary_data)
        f.close()

    @staticmethod
    def write_json_to_file(dict: dict, filename: str):
        """
        Writes the provided 'dict' as a json into the 'filename'.

        @param
            **dict**
            Python dictionary that will be stored as a json.

            **filename**
            File path in which we are going to store the information.
        """
        if dict == None:
            return None
        
        if not filename:
            return None
        
        return FileWriter.write_file(json.dumps(dict, indent = 4), filename)

    @staticmethod
    def write_file(text: str, output_filename: str):
        """
        Writes the provided 'text' in the 'filename' file. It replaces the previous content
        if existing.
        """
        if not text:
            return None
        
        if not output_filename:
            return None

        f = open(output_filename, 'w', encoding = 'utf8')
        f.write(text)
        f.close()

        return output_filename

    @staticmethod
    def write_file_by_chunks_from_response(response: Response, output_filename: str):
        """
        Iterates over the provided 'response' and writes its content
        chunk by chunk in the also provided 'output_filename'.

        TODO: If you find a better way to handle this you are free to
        create new methods and move them into a new file.
        """
        if not response:
            return None
        
        if not output_filename:
            return None
        
        CHUNK_SIZE = 32768

        # TODO: Make this method work with a common Iterator parameter
        # and not an specific response, please
        with open(output_filename, 'wb') as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return output_filename


from yta_general_utils.file.enums import FileType
from moviepy import VideoFileClip, AudioFileClip

import os


def sanitize_filename(filename: str):
    """
    This method checks the provided 'filename' and turns any 
    backslash character into a '/' (slash) one, returning the
    new string.
    """
    if '\\' in filename:
        filename = filename.replace('\\', '/')

    return filename

def get_filename(filename):
    """
    This method returns the filename, avoiding the path, of
    the provided 'filename'. This method includes the extension
    if existing.
    """
    aux = sanitize_filename(filename).split('/')

    return aux[len(aux) - 1]

def filename_is_type(filename: str, type: FileType):
    """
    Checks if the provided 'filename' is a valid filename and if 
    its type is the given 'type' or not (based on the extension).
    This method will return True if the 'filename' is valid and
    belongs to the provided 'type', or False if not. It wil raise
    a Exception if something is bad formatted or missing.
    """
    if not filename:
        raise Exception('No "filename" provided.')
    
    if not isinstance(filename, str):
        raise Exception('The "filename" provided is not a str.')

    type = FileType.to_enum(type)
    
    extension = get_file_extension(filename)
    if not extension:
        raise Exception('Provided "filename" has no valid extension to be checked.')
    
    return type.is_accepted_value(extension)

# TODO: Rename these methods below to 'filename' instead of 'file'
# TODO: Implement the use of our RegularExpression enums
def file_is_type(filename: str, type: FileType):
    """
    Checks if the provided 'filename' is a valid filename and if 
    its type is the given 'type' or not (based on the extension).
    This method will return True if the 'filename' is valid, 
    belongs to the provided 'type' and is a valid file (it will
    be checked by trying to instantiate the file as real), or 
    False if not. It wil raise an Exception if something is bad 
    formatted or missing.
    """
    if not filename:
        raise Exception('No "filename" provided.')
    
    if not isinstance(filename, str):
        raise Exception('The "filename" provided is not a str.')

    type = FileType.to_enum(type)
    
    extension = get_file_extension(filename)
    if not extension:
        raise Exception('Provided "filename" has no valid extension to be checked.')
    
    if not filename_is_type(filename, type):
        return False
    
    # TODO: I should do this with another library to
    # avoid 'moviepy' dependency
    if type == FileType.VIDEO:
        try: 
            VideoFileClip(filename)
        except:
            return False
    elif type == FileType.AUDIO:
        try:
            AudioFileClip(filename)
        except:
            return False
        
# TODO: Maybe move this below to a 'file_checker.py'?
def file_has_extension(filename):
    if get_file_extension(filename):
        return True
    
    return False

def ensure_file_extension(filename: str, extension: str):
    """
    This method returns the provided 'filename' with the
    also provided 'extension', making sure that this
    extension is used in the filename.

    This method is useful when you need a file to have an
    specific extension.
    """
    # TODO: Please, improve this method...
    if extension is None or not isinstance(extension, str) or not extension:
        raise Exception('No valid "extension" provided.')
    
    if '.' in extension:
        extension = extension.replace('.', '')

    # TODO: Ensure extension is a valid file extension
    filename_extension = get_file_extension(filename)
    if not filename_extension:
        filename += f'.{extension}'
    elif filename_extension != extension:
        filename = replace_file_extension(filename, extension)

    return filename

def get_file_extension(filename: str):
    """
    Returns the extension of the provided 'filename' that
    doesn't include the '.'.

    If you provide '.png' as 'filename' it will return
    a False. If the filename has no extension, it will 
    return None.
    """
    # TODO: Is this (https://stackoverflow.com/a/49689414) better (?)
    filename, ext = os.path.splitext(filename)

    if '.' in ext:
        ext = ext.replace('.', '')

    return ext

def replace_file_extension(filename: str, extension: str):
    """
    Replaces the original 'filename' extension with the provided
    'extension'.
    """
    # TODO: Improve this code below by raising exceptions
    if not filename:
        return None
    
    if not extension:
        return None
    
    if not '.' in extension:
        extension = '.' + extension

    root, ext = os.path.splitext(extension)
    
    if root and ext:
        # Provided 'extension' is not valid
        return None

    root, ext = os.path.splitext(filename)
    
    return root + extension

def get_file_filename_without_extension(filename: str):
    """
    This method receives a full 'filename' that could be
    an absolute or relative path (including slashes) and
    returns only the filename part (excluding the
    extension and path).

    (!) Passing '.png' or similar as 'filename' parameter
    will fail.

    TODO: Maybe check this condition above (?) but what
    about '.gitignore' file?
    """
    if not filename:
        return None
    
    filename = get_filename(filename)
    filename, ext = os.path.splitext(filename)

    return filename


from math import gcd


# TODO: Maybe I can refactor this with the new method below
def resize_to_fit_on_region(element_size: tuple, region_size: tuple):
    """
    This method calculates the size that the 'element_size' must be
    resized to in order to fit the provided 'region_size'. Fitting 
    means that the element will cover the whole region.

    This method is useful to resize images or videos that will be
    placed behind alpha (transparent) areas of images so they will
    fit perfectly.

    This method returns a group of two values that are the expected
    element width and height (w, h).
    """
    # TODO: Check that the format is ok and the values are also ok
    if len(element_size) != 2 or len(region_size) != 2:
        raise Exception('The provided "element_size" and "region_size" parameters are not tuples of (w, h) values.')

    element_width, element_height = element_size
    region_width, region_height = region_size

    great_common_divisor = gcd(element_width, element_height)
    step_x = element_width / great_common_divisor
    step_y = element_height / great_common_divisor

    # Make sure they are even numbers to be able to move at least
    # one pixel on each side
    if step_x % 2 != 0 or step_y % 2 != 0:
        step_x *= 2
        step_y *= 2

    # If element is larger than region, we need to make it smaller.
    # In any other case, bigger
    if element_width > region_width and element_height > region_height:
        step_x = -step_x
        step_y = -step_y
    
    do_continue = True
    tmp_size = [element_width, element_height]
    while (do_continue):
        tmp_size = [tmp_size[0] + step_x, tmp_size[1] + step_y]

        if step_x < 0 and (tmp_size[0] < region_width or tmp_size[1] < region_height):
            # The previous step had the right dimensions
            tmp_size[0] += abs(step_x)
            tmp_size[1] += abs(step_y)
            do_continue = False
        elif step_x > 0 and (tmp_size[0] > region_width and tmp_size[1] > region_height):
            # This step is ok
            do_continue = False

    return tmp_size[0], tmp_size[1]


from yta_general_utils.coordinate.coordinate import Coordinate


class Region:
    """
    Class to represent a region built by two
    coordinates, one in the top left corner
    and another one in the bottom right corner.
    """

    top_left: Coordinate = None
    bottom_right: Coordinate = None
    # TODO: Do we actually need them (?)
    coordinates: list[Coordinate] = None
    width: int = None
    height: int = None

    def __init__(self, top_left_x: int, top_left_y: int, bottom_right_x: int, bottom_right_y: int, coordinates: list[Coordinate]):
        self.top_left = Coordinate(top_left_x, top_left_y)
        self.bottom_right = Coordinate(bottom_right_x, bottom_right_y)
        self.coordinates = coordinates
        self.width = self.bottom_right.x - self.top_left.x
        self.height = self.bottom_right.y - self.top_left.y

    @property
    def size_to_fit(self) -> tuple[int, int]:
        """
        Size that guarantees any element fits the region.
        This includes a 1% of increase to make sure no
        black pixels are shown in the borders.
        """
        return (
            int(self.width * 1.01),
            int(self.height * 1.01)
        )
    
    @property
    def center(self) -> tuple[int, int]:
        """
        The center position of the region represented by
        a tuple (x, y).
        """
        return (
            int(self.bottom_right.x - self.width / 2),
            int(self.bottom_right.y - self.height / 2)
        )
    



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from typing import Union

import time


class ChromeScrapper:
    """
    A class that wraps and simplify the functionality of
    the Google Chrome scrapper, useful to interact with
    websites and navigate through them interacting with
    their elements being able to perform complex actions
    such as downloading files, sending forms, etc.
    """
    gui = True
    ad_blocker = True
    disable_popups_and_cookies = True
    additional_options = []
    max_page_load_waiting_time = 10
    options = None
    # TODO: Maybe make 'driver' private attribute in ChromeScrapper to
    # avoid exposing it
    driver = None

    @property
    def active_element(self):
        """
        Get the current active element by running the
        'driver.switch_to.active_element' command.
        """
        return self.driver.switch_to.active_element if self.driver is not None else None

    def __init__(self, gui: bool = False, ad_blocker: bool = True, disable_popups_and_cookies: bool = True, additional_options = [], max_page_load_waiting_time: float = 10):
        # TODO: Make more customizable accepting ad_blocker extension name
        # and stuff like that
        self.gui = gui
        self.ad_blocker = ad_blocker
        self.disable_popups_and_cookies = disable_popups_and_cookies
        self.additional_options = additional_options
        self.max_page_load_waiting_time = max_page_load_waiting_time

        self.__init_options()
        self.__init_driver()

    def __del__(self):
        try:
            self.driver.close()
        finally:
            self.driver = None

    def __init_options(self):
        """
        Initializes the Google Chrome options and returns them.
        """
        if not self.options:
            # TODO: This must be dynamic and/or given by user
            CHROME_EXTENSIONS_ABSOLUTEPATH = 'C:/Users/dania/AppData/Local/Google/Chrome/User Data/Profile 2/Extensions/'
            # TODO: Extensions versions are updated, so check below line
            AD_BLOCK_ABSOLUTEPATH = CHROME_EXTENSIONS_ABSOLUTEPATH + 'cjpalhdlnbpafiamejdnhcphjbkeiagm/1.59.0_0'
            # TODO: What if 'import undetected_chromedriver.v2 as uc'? Try it

            options = Options()
            
            # TODO: Make this a dynamic option that can be passed through __init__
            option_arguments = ['window-size=1920,1080']
            # option_arguments = ['--start-maximized']

            if len(self.additional_options) > 0:
                for additional_option in self.additional_options:
                    option_arguments.append(additional_option)

            if not self.gui:
                option_arguments.append('--headless=new')

            if self.ad_blocker:
                # This loads the ad block 'uBlock' extension that is installed in my pc
                option_arguments.append('load-extension=' + AD_BLOCK_ABSOLUTEPATH)
            
            # Load user profile
            option_arguments.append('user-data-dir=C:/Users/dania/AppData/Local/Google/Chrome/User Data/Profile 2')

            # Ignore certs
            option_arguments.append('--ignore-certificate-errors')
            option_arguments.append('--ignore-ssl-errors')
            option_arguments.append('--ignore-certificate-errors-spki-list')

            for argument in option_arguments:
                options.add_argument(argument)

            if self.disable_popups_and_cookies:
                # TODO: Separate this into specific options, not all together. One is for cookies,
                # another one is for popups... Separate them, please
                # This disables popups, cookies and that stuff
                options.add_experimental_option('prefs', {
                    'excludeSwitches': ['enable-automation', 'load-extension', 'disable-popup-blocking'],
                    'profile.default_content_setting_values.automatic_downloads': 1,
                    'profile.default_content_setting_values.media_stream_mic': 1
                })

            self.options = options

        return self.options
    
    def __init_driver(self):
        """
        Initializes the Google Chrome driver and returns it.
        """
        if not self.driver:
            self.driver = webdriver.Chrome(options = self.options)
            # We force to start a new session due to some problems
            #self.driver.start_session({})

        return self.driver
    
    def go_to_web_and_wait_util_loaded(self, url: str):
        """
        Navigates to the provided 'url' and checks continuously
        if the page has been loaded or not. It will wait for
        'max_page_load_waiting_time' seconds set when the
        ChromeScrapper object was created.

        This method will return True in the moment the page is 
        loaded.
        """
        if not url:
            raise Exception('No "url" provided.')
    
        CHECK_TIME = 0.25

        try:
            self.driver.get(url)

            page_state = 'loading'
            cont = 0
            while page_state != 'complete' and cont < (self.max_page_load_waiting_time / CHECK_TIME):
                time.sleep(CHECK_TIME)
                page_state = self.driver.execute_script('return document.readyState;')
                cont += 1

            if page_state == 'loading':
                return False
            
            return True

        except:
            self.driver.close()
            self.driver = None

    def wait(self, seconds: float):
        """
        Waits for the provided 'time' seconds. Value must be between
        0 and 60.
        """
        if not seconds:
            raise Exception('No "seconds" provided.')
        
        if seconds < 0:
            raise Exception('Provided "seconds" must be greater than 0.')
        
        if seconds > 60:
            seconds = 60

        time.sleep(seconds)

    def press_ctrl_letter(self, letter: str = 'c'):
        """
        Holds the Ctrl key down, presses the provided 'letter' and 
        releases the Ctrl key. This is useful for Ctrl+C, Ctrl+V
        combinations.
        """
        # TODO: Check if letter is valid
        if not letter:
            raise Exception('No "letter" provided.')
        
        letter = letter[0]

        ActionChains(self.driver).key_down(Keys.CONTROL).send_keys(letter).key_up(Keys.CONTROL).perform()

    def press_ctrl_letter_on_element(self, letter: str, element: WebElement):
        """
        Presses the Ctrl + provided 'letter' in the also provided
        'element'. Useful to paste text into text elements.
        """
        if not letter:
            raise Exception('No "letter" provided.')
        
        if not element:
            raise Exception('No "element" provided.')
        
        letter = letter[0]

        element.send_keys(Keys.CONTROL, letter)

    def press_ctrl_c(self):
        self.press_ctrl_letter('c')

    def press_ctrl_c_on_element(self, element: WebElement):
        self.press_ctrl_letter_on_element('c', element)

    def press_ctrl_x(self):
        self.press_ctrl_letter('x')

    def press_ctrl_x_on_element(self, element: WebElement):
        self.press_ctrl_letter_on_element('x', element)

    def press_ctrl_v(self):
        self.press_ctrl_letter('v')

    def press_ctrl_v_on_element(self, element: WebElement):
        self.press_ctrl_letter_on_element('v', element)

    def press_ctrl_a(self):
        self.press_ctrl_letter('a')

    def press_ctrl_a_on_element(self, element: WebElement):
        self.press_ctrl_letter_on_element('a', element)

    def press_key_x_times(self, key: Keys, times: int):
        """
        Presses the provided 'key' 'times' times one behind
        the other one. This method is useful to use TAB,
        ENTER or keys like that a lot of times.
        """
        # TODO: Check times is valid
        actions_chain = ActionChains(self.driver)
        for i in range(times):
            actions_chain.send_keys(key)
        actions_chain.perform()

    def find_element_by_id(self, id: str, element: Union[WebElement, None] = None):
        """
        This method returns the first WebElement found with
        the provided 'id' or None if not found.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.
        """
        elements = self.find_elements_by_id(id, element)

        if not elements:
            return None
        
        return elements[0]
    
    def find_element_by_id_waiting(self, id: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 
        'element_type' with also provided 'id' is visible and 
        returns it if it becomes visible in the 'time' seconds 
        of waiting. It returns None if not.
        """
        if not id:
            raise Exception('No "id" provided.')
        
        return self.__find_element_by_waiting(By.ID, id, time)

    def find_elements_by_id(self, id: str, element: Union[WebElement, None] = None):
        """
        This method returns an array containing the WebElements
        found with the provided 'id'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.
        """
        if not id:
            raise Exception('No "id" provided.')
        
        if element:
            return element.find_elements(By.ID, id)
        
        return self.driver.find_elements(By.ID, id)

    def find_element_by_text(self, element_type: str, text: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[contains(text(), 'text')]' 
        structure to find the element, useful for buttons that 
        have 'Save' text or things similar.

        You can use 'element_type' = 'button' and 'text' =
        'Guardar' to find the elements like this one:
        <button>Guardar</button>

        You can also use the wildcard '*'  to find any type of
        element with the specific provided 'text'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns the first found WebElement if 
        existing or None if not found.
        """
        elements = self.find_elements_by_text(element_type, text, element)
        
        if len(elements) > 0:
            return elements[0]
        
        return None

    def find_element_by_text_waiting(self, element_type: str, text: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 
        'element_type' and 'text' is visible and returns it if 
        it becomes visible in the 'time' seconds of waiting. It 
        returns None if not.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not text:
            raise Exception('No "text" provided.')

        return self.__find_element_by_waiting(By.XPATH, "//" + element_type + "[contains(text(), '" + text + "')]", time)
    
    def find_elements_by_text(self, element_type: str, text: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[contains(text(), 'text')]' 
        structure to find the element, useful for buttons that 
        have 'Save' text or things similar.

        You can use 'element_type' = 'button' and 'text' =
        'Guardar' to find the elements like this one:
        <button>Guardar</button>

        You can also use the wildcard '*'  to find any type of
        element with the specific provided 'text'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns an array with all the found elements
        or empty if not found.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not text:
            raise Exception('No "text" provided.')
        
        if element:
            return element.find_elements(By.XPATH, "//" + element_type + "[contains(text(), '" + text + "')]")

        return self.driver.find_elements(By.XPATH, "//" + element_type + "[contains(text(), '" + text + "')]")
    
    def find_element_by_class(self, element_type: str, class_str: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[contains(@class, 'class_str')]' 
        structure to find the element, useful for divs with a
        specific class or similar.

        You can use 'element_type' = 'div' and 'class_str' =
        'container-xl' to find the elements like this one:
        <div class='container-xl'>content</div>

        You can also use the wildcard '*'  to find any type of
        element with the specific provided 'class_str'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns the first found WebElement if 
        existing or None if not found.
        """
        elements = self.find_elements_by_class(element_type, class_str, element)
        
        if len(elements) > 0:
            return elements[0]
        
        return None
    
    def find_element_by_class_waiting(self, element_type: str, class_str: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 
        'element_type' and class is visible and returns it if it
        becomes visible in the 'time' seconds of waiting. It 
        returns None if not.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not class_str:
            raise Exception('No "class_str" provided.')
        
        return self.__find_element_by_waiting(By.XPATH, "//" + element_type + "[contains(@class, '" + class_str + "')]", time)

    def find_elements_by_class(self, element_type: str, class_str: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[contains(@class, 'class_str')]' 
        structure to find the element, useful for divs with a
        specific class or similar.

        You can use 'element_type' = 'div' and 'class_str' =
        'container-xl' to find the elements like this one:
        <div class='container-xl'>content</div>

        You can also use the wildcard '*'  to find any type of
        element with the specific provided 'class_str'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns an array with all the found elements
        or empty if not found.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not class_str:
            raise Exception('No "class_str" provided.')

        if element:
            return element.find_elements(By.XPATH, "//" + element_type + "[contains(@class, '" + class_str + "')]")
        
        return self.driver.find_elements(By.XPATH, "//" + element_type + "[contains(@class, '" + class_str + "')]")

    def find_element_by_custom_tag(self, element_type: str, custom_tag: str, custom_tag_value: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[@custom-tag='custom_value')]' 
        structure to find the element, useful for divs with a
        specific tag or similar.

        You can use the wildcard '*'  to find any type of
        element with the specific provided 'custom_tag'.

        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns the first found WebElement if 
        existing or None if not found.
        """
        elements = self.find_elements_by_custom_tag(element_type, custom_tag, custom_tag_value, element)
        
        if len(elements) > 0:
            return elements[0]
        
        return None

    def find_element_by_custom_tag_waiting(self, element_type: str, custom_tag: str, custom_tag_value: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 
        'element_type' and custom tag is visible and returns it 
        if it becomes visible in the 'time' seconds of waiting.
        It returns None if not.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not custom_tag:
            raise Exception('No "class_str" provided.')
        
        tag = '@' + custom_tag
        if custom_tag_value:
            tag += "='" + custom_tag_value + "'"

        return self.__find_element_by_waiting(By.XPATH, "//" + element_type + "[" + tag + "]", time)

    def find_elements_by_custom_tag(self, element_type: str, custom_tag: str, custom_tag_value: str, element: Union[WebElement, None] = None):
        """
        This method uses the 'By.XPATH' finding elements method
        with the '//element_type[@custom-tag='custom_value')]' 
        structure to find the element, useful for divs with a
        specific tag or similar.

        You can use the wildcard '*'  to find any type of
        element with the specific provided 'custom_tag'.
        
        If you provide the 'element' parameter, the search will 
        be in that element instead of the whole web page.

        This method returns an array with all the found elements
        or empty if not found.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        if not custom_tag:
            raise Exception('No "class_str" provided.')
        
        # @custom-tag or @custom-tag='something'
        tag = '@' + custom_tag
        if custom_tag_value:
            tag += "='" + custom_tag_value + "'"

        if element:
            return element.find_elements(By.XPATH, "//" + element_type + "[" + tag + "]")

        return self.driver.find_elements(By.XPATH, "//" + element_type + "[" + tag + "]")

    def find_element_by_element_type(self, element_type: str, element: Union[WebElement, None] = None):
        """
        Returns the elements found with the provided 'element_type' tag, that
        is the tag name ('span', 'div', etc.). If you provide the 'element' 
        parameter, the search will be in that element instead of the whole
        web page.
        """
        # TODO: Document
        elements = self.find_elements_by_element_type(element_type, element)
        
        if len(elements) > 0:
            return elements[0]
        
        return None
    
    def find_element_by_element_type_waiting(self, element_type: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 
        'element_type' tag is visible and returns it if it becomes
        visible in the 'time' seconds of waiting. It returns None
        if not.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        return self.__find_element_by_waiting(By.TAG_NAME, element_type, time)
        
    def find_elements_by_element_type(self, element_type: str, element: Union[WebElement, None] = None, only_first_level: bool = False):
        """
        Returns the web elements with the provided 'element_type' tag. 
        This method will search in the 'element' if provided, or in 
        the whole web page if not. It will look for elements only on
        the first level y 'only_first_level' is True, or in any level
        if False.

        If 'only_first_level' is True, this will look only in the 
        first child level, horizontally, so a child of a child tag
        won't be returned.
        """
        if not element_type:
            raise Exception('No "element_type" provided.')
        
        object = self.driver
        if element:
            object = element

        if only_first_level:
            # This will only look on first children horizontally
            return object.find_elements(By.XPATH, './' + element_type)
                                         
        return object.find_elements(By.TAG_NAME, element_type)
    
    def find_element_by_xpath_waiting(self, xpath: str, time: int = 30):
        """
        Waits until the WebElement corresponding to the provided 'xpath'
        is visible and returns it if it becomes visible in the 'time' 
        seconds of waiting. It returns None if not.
        """
        return self.__find_element_by_waiting(By.XPATH, xpath, time)
    
    # TODO: Create 'find_elements_by_xpath'
    
    def __find_element_by_waiting(self, by: By, by_value: str, time: int = 30):
        """
        Internal method to simplify the waiting for element until visible
        code reusability.
        """
        wait = WebDriverWait(self.driver, time)
        element = wait.until(EC.visibility_of_element_located((by, by_value)))

        if not element:
            return None
        
        return element
    
    # TODO: When an element is hidden and you cannot interact you
    # can change the style.display
    # driver.execute_script("arguments[0].style.display = 'block';", field)

    def set_file_input(self, element: WebElement, abspath: str):
        """
        Sends the file to a 'type=file' input web element. The 
        provided 'abspath' must be the absolute path to the file
        you want to send.
        """
        if not element:
            raise Exception('No file input "element" provided.')
        
        if not abspath:
            raise Exception('No file "abspath" provided.')
        
        # TODO: Check that it is a valid abspath

        element.send_keys(abspath)
    
    def get_current_page_y_offset(self):
        """
        Returns the current page Y axis offset by executing the
        'return window.pageYOffset' script. This means the amount
        of pixels moved from the origin (top). An offset of 50 
        pixels means that has been a scroll down of 50 pixels.
        The minimum value is 0 when on top of the web page.
        """
        return self.driver.execute_script("return window.pageYOffset")

    def scroll_down(self, pixels: int):
        """
        Scrolls down the web page the amount of 'pixels' provided
        as parameter from the current position. This method will
        make a passive waiting until the new position is reached.
        """
        # TODO: Check if 'pixels' is valid

        pixels = abs(pixels)
        pixels += self.get_current_page_y_offset()

        self.driver.execute_script('window.scrollTo(0, ' + str(pixels) + ')')

        # We wait until movement is completed
        waiting_times = 300
        while (self.get_current_page_y_offset() != pixels and waiting_times > 0):
            self.wait(0.1)
            waiting_times -= 1

    def scroll_up(self, pixels: int):
        """
        Scrolls up the web page the amount of 'pixels' provided
        as parameter from the current position. This method will
        make a passive waiting until the new position is reached.
        """
        # TODO: Check if 'pixels' is valid

        pixels = abs(pixels)
        current_y = self.get_current_page_y_offset()
        if current_y - pixels > 0:
            pixels = current_y - pixels
        else:
            pixels = 0

        self.driver.execute_script('window.scrollTo(0, ' + str(pixels) + ')')

        waiting_times = 300
        while (self.get_current_page_y_offset() != pixels and waiting_times > 0):
            self.wait(0.1)
            waiting_times -= 1

    def scroll_to_element(self, element: WebElement):
        """
        This method scrolls to 50 pixels above the web element
        to make sure it is on the middle of the web page. This
        is very useful to take screenshots. This method will
        make a passive waiting until the new position is 
        reached.
        """
        element_y = element.location['y']
        y = 0

        if element_y > 50:
            y = element_y - 50

        self.execute_script('window.scrollTo(0, ' + str(y) + ')')

        # We wait until movement is completed
        waiting_times = 300
        while (self.get_current_page_y_offset() != y and waiting_times > 0):
            self.wait(0.1)
            waiting_times -= 1

    def get_page_height(self):
        """
        Returns the current page height by executing the
        'return document.body.scrollHeight' script. This means
        the amount of pixels from top to bottom.
        """
        return self.driver.execute_script("return document.body.scrollHeight")

    def screenshot(self, output_filename: str = None):
        """
        Takes a screenshot of the whole page and returns the 
        it as binary data if no 'output_filename' provided. 
        If 'output_filename' is provided, it will be stored 
        locally with that name.
        """
        # TODO: Make this method return the image as binary
        # data always and store if 'output_filename' is 
        # provided, but you cannot do the 'save_screenshot'
        # twice because the webpage can change in the time
        # elapsed between both screenshots.
        if not output_filename:
            return self.driver.get_screenshot_as_png()
        
        self.driver.save_screenshot(output_filename)

        return output_filename
    
    def screenshot_element(self, element: WebElement, output_filename: str = None):
        """
        Takes a screenshot of the provided 'element', that
        means that only the area occupied by that element
        is shown in the screenshot, and returns it as 
        binary data if no 'output_filename' provided or
        will be stored locally if provided.

        Any element of the web page that is over the 
        element will appear in the screenshot blocking it.

        This method will return the 'output_filename' if it
        was provided, so the file has been stored locally,
        or a dict containing 'size' (width, height) and 
        'data' fields.
        """
        # TODO: Make this method return the image as binary
        # data always and store if 'output_filename' is 
        # provided, but you cannot do the 'save_screenshot'
        # twice because the webpage can change in the time
        # elapsed between both screenshots.
        if not element:
            raise Exception('No "element" provided.')
        
        if not output_filename:
            return element.screenshot_as_png
        
        element.screenshot(output_filename)

        return output_filename

    def screenshot_web_page(self, url: str):
        """
        This method will take screenshots of the whole
        web page. It will do in the current page if
        no 'url' is provided, or will navigate to the
        provided 'url' and do it in that one.
        """
        if url:
            self.go_to_web_and_wait_util_loaded(url)

        # TODO: Screenshot the whole page
        FPS = 60
        # Maybe this should be a parameter
        # TODO: We want to make screenshots for a video
        # so depending on 'duration' it will be slower
        # or more dynamic. This method need testing
        duration = 5

        # TODO: Maybe we want to scroll more, or maybe
        # we should pass this a parameter to make it
        # more customizable
        page_height = self.get_page_height()
        if page_height > 1000:
            page_height = 1000

        screenshots = []
        number_of_screenshots = int(duration * FPS)
        window_size = self.driver.get_window_size()

        # TODO: How much should we scroll?
        new_height = 0
        for i in range(number_of_screenshots):
            screenshots.append(self.driver.get_screenshot_as_png())
            height = self.get_current_page_y_offset()
            new_height += page_height / number_of_screenshots
            # We scroll down the difference
            self.scroll_down(new_height - height)

        # TODO: What if we end before

        return screenshots

    def execute_script(self, script: str, *args):
        """
        Executes the provided 'script' synchronously.
        """
        if not script:
            raise Exception('No "script" provided.')
        
        self.driver.execute_script(script, *args)

    def remove_element(self, element: WebElement):
        # TODO: Document
        if not element:
            raise Exception('No "element" provided.')
        
        self.execute_script('arguments[0].remove()', element)
    
    def set_element_width(self, element: WebElement, width: int):
        # TODO: Document
        if not element:
            raise Exception('No "element" provided.')
        
        if not width:
            raise Exception('No "width" provided.')
        
        if width < 0:
            raise Exception('Parameter "width" must be greater than 0.')
        
        # TODO: Do more checkings (is width number and int?)
        
        self.execute_script('arguments[0].style = "width: ' + str(width) + 'px;"', element)

    def set_element_style(self, element: WebElement, style: str):
        # TODO: Document
        if not element:
            raise Exception('No "element" provided.')
        
        if not style:
            raise Exception('No "style" provided.')
        
        self.execute_script('arguments[0].style = "' + style + '"', element)

    def set_element_attribute(self, element: WebElement, attribute: str, value: str):
        # TODO: Document
        if not element:
            raise Exception('No "element" provided.')
        
        if not attribute:
            raise Exception('No "attribute" provided.')
        
        if not value:
            raise Exception('No "value" provided.')
        
        self.execute_script('arguments[0].setAttribute("' + attribute + '", "' + value + '")', element)

    def set_element_inner_text(self, element: WebElement, inner_text: str):
        """
        Sets the provided 'inner_text' in the also provided 'element'
        by applying the script 'element.innerText = "inner_text";'.
        """
        if not element:
            raise Exception('No "element" provided.')
        
        if not inner_text:
            raise Exception('No "inner_text" provided.')

        self.execute_script('arguments[0].innerText = "' + str(inner_text) + '";', element)

    def get_page_size(self):
        """
        Returns a tuple (width, height) containing the
        size of the current web page (attending to the
        screen used).
        """
        size = self.driver.get_window_size()

        return (size['width'], size['height'])
    
    def set_page_size(self, width: int = 1920, height: int = 1080):
        """
        This method resizes the web navigator to the provided
        width and height. It is useful to take screenshots from
        webpages or to validate different screen sizes.
        """
        if not width:
            raise Exception('No "width" provided.')
        
        if not height:
            raise Exception('No "height" provided.')
        
        # TODO: Do more checkings

        self.driver.set_window_size(width, height)

    def add_to_clipboard(self, text: str):
        """
        Adds the provided 'text' to the clipboard to be able to 
        paste it. This method will create a 'textarea' element,
        write the provided 'text' and copy it to the web
        scrapper clipboard.
        """
        TEXT_AREA_ID = 'textarea_to_copy_912312' # a random one
        # I create a new element to put the text, copy it and be able to paste
        js_code = "var p = document.createElement('textarea'); p.setAttribute('id', '" + TEXT_AREA_ID + "'); p.value = '" + text + "'; document.getElementsByTagName('body')[0].appendChild(p);"
        self.execute_script(js_code)

        # Focus on textarea
        textarea = self.find_element_by_id(TEXT_AREA_ID)
        textarea.click()
        self.press_ctrl_a_on_element(textarea)
        self.press_ctrl_c_on_element(textarea)
        # TODO: I can use 'textarea.send_keys(Keys.CONTROL, 'c') to copy, validate
        # actions.key_down(Keys.CONTROL).send_keys('A').key_up(Keys.CONTROL).perform()
        # actions.key_down(Keys.CONTROL).send_keys('C').key_up(Keys.CONTROL).perform()

        # Remove the textarea, it is no longer needed
        js_code = "var element = document.getElementById('" + TEXT_AREA_ID + "'); element.parentNode.removeChild(element);"
        # TODO: Update this with the new version
        self.execute_script(js_code)

    # TODO: Maybe automate some 'execute_javascript' to change
    # 'innerHTML' and that stuff (?)



# TODO: Remove all this below when all refactored
    


def download_fake_call_image(name, output_filename):
    # TODO: Move this to the faker
    URL = 'https://prankshit.com/fake-iphone-call.php'

    try:
        driver = start_chrome()
        go_to_and_wait_loaded(driver, URL)

        inputs = driver.find_elements(By.TAG_NAME, 'input')
        name_textarea = driver.find_element(By.TAG_NAME, 'textarea')

        #operator_input = inputs[4]
        #hour_input = inputs[5]

        name_textarea.clear()
        name_textarea.send_keys(name)

        image = driver.find_element(By.XPATH, '//div[contains(@class, "modal-content tiktok-body")]')
        image.screenshot(output_filename)
    finally:
        driver.close()

# Other fake generators (https://fakeinfo.net/fake-twitter-chat-generator) ad (https://prankshit.com/fake-whatsapp-chat-generator.php)
def download_discord_message_image(text, output_filename):
    URL = 'https://message.style/app/editor'

    try:
        driver = start_chrome()
        go_to_and_wait_loaded(driver, URL)
        
        time.sleep(3)

        clear_embed_button = driver.find_element(By.XPATH, "//button[contains(text(), 'Clear Embeds')]")
        clear_embed_button.click()

        time.sleep(3)

        input_elements = driver.find_elements(By.TAG_NAME, 'input')
        username_input = input_elements[3]
        avatar_url_input = input_elements[4]

        username_input.clear()
        username_input.send_keys('botsito')

        avatar_url_input.clear()
        avatar_url_input.send_keys('https://cdn.pixabay.com/photo/2016/11/18/23/38/child-1837375_640.png')

        textarea_input = driver.find_element(By.TAG_NAME, 'textarea')
        textarea_input.clear()
        textarea_input.send_keys(text)

        time.sleep(3)

        # get element div class='discord-message'
        discord_message = driver.find_element(By.XPATH, "//div[contains(@class, 'discord-message')]")
        discord_message.screenshot(output_filename)
    finally:
        driver.close()



def test_download_piano_music():
    # TODO: End this, to make it download music generated by this AI
    try:
        options = Options()
        #option_arguments = ['--start-maximized', '--headless=new']
        option_arguments = ['--start-maximized']
        for argument in option_arguments:
            options.add_argument(argument)

        driver = webdriver.Chrome(options = options)
        driver.get('https://huggingface.co/spaces/mrfakename/rwkv-music')
        wait = WebDriverWait(driver, 30)
        download_button_element = wait.until(EC.visibility_of_element_located((By.XPATH, '//*[@data-testid=\"checkbox\"]')))

        time.sleep(5)
        actions = ActionChains(driver)
        for i in range(11):
            actions.send_keys(Keys.TAB)
        actions.perform()
        actions.send_keys(Keys.SPACE)
        actions.perform()

        input_number_element = driver.find_elements_by_xpath('//*[@data-testid="number-input"]')[0]
        input_number_element.send_keys(14286)

        actions = ActionChains(driver)
        actions.send_keys(Keys.TAB)
        actions.perform()

        time.sleep(1)

        driver.execute_script('window.scrollTo(0, 1000)')

        time.sleep(4)

        download_button_element = wait.until(EC.visibility_of_element_located((By.XPATH, '//*[@title=\"Download\"]')))
        #download_button_element = driver.find_elements_by_xpath('//*[@title="Download"]')[0]
        download_button_element.click()

    finally:
        driver.close()


# TODO: Move this method below to the main class
def get_redicted_url(url, expected_url = None, wait_time = 5):
    """
    Navigates to the provided url and waits for a redirection. This method will wait
    until the 'expected_url' is contained in the new url (if 'expected_url' parameter
    is provided), or waits 'wait_time' seonds to return the current_url after that.
    """
    redirected_url = ''

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Remove this line below for debug
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        wait = WebDriverWait(driver, 10)

        if not expected_url:
            time.sleep(wait_time)
        else:
            wait.until(EC.url_contains(expected_url))

        redirected_url = driver.current_url
    finally:
        driver.close()

    return redirected_url


def get_youtube_summary(video_id):
    """
    Searchs into 'summarize.tech' web to obtain the summary of the video with the 
    provided 'video_id'. This method returns the summary in English, as it is 
    provided by that website.
    """
    url = 'https://www.summarize.tech/www.youtube.com/watch?v=' + video_id

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Remove this line below for debug
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        summary = driver.find_element_by_tag_name('section').find_element_by_tag_name('p').get_attribute('innerText')
    finally:
        driver.close()

    return summary

def google_translate(text, input_language = 'en', output_language = 'es'):
    url = 'https://translate.google.com/?hl=es'
    """
    https://translate.google.com/?hl=es&sl=en&tl=es&text=Aporta%20una%20unidad%20de%20traducci%C3%B3n%20(segmento%20y%20traducci%C3%B3n)%20en%20alg%C3%BAn%20par%20de%20idiomas%20a%20MyMemory.%0ASin%20especificar%20ning%C3%BAn%20par%C3%A1metro%20clave%2C%20la%20contribuci%C3%B3n%20est%C3%A1%20disponible%20para%20todos%20(%C2%A1Gracias!).&op=translate

    https://translate.google.com/?hl=es&tab=TT&sl=en&tl=es&op=translate

    https://translate.google.com/?hl=es&tab=TT&sl=en&tl=es&text=La%20%C3%BAnica%20forma%20de%20saberlo%20es%20lo%20que%20t%C3%BA%20digas&op=translate
    """

    url = 'https://translate.google.com/?hl=' + output_language + '&tab=TT&sl=' + input_language + '&tl=' + output_language + '&text=' + text + '&op=translate'

    translation = ''

    # TODO: Verify that this below is working
    driver = ChromeScrapper()
    driver.go_to_web_and_wait_util_loaded(url)
    translation = driver.find_element_by_xpath_waiting('//*[@jscontroller="JLEx7e"]', 10).get_attribute('innerText')
    # TODO: Maybe make a method to obtain the inner text from a WebElement (?)
    return translation

    # TODO: Refactor this as we have a better chrome driver lib now
    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Comment this line below for debug (enables GUI)
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        tries = 0
        while True:
            if tries < 20:
                try:
                    # We try until it doesn't fail (so we have the text)
                    # TODO: This 'jscontroller' changes from time to time, pay atention
                    translation = driver.find_elements('xpath', '//*[@jscontroller="JLEx7e"]')[0].get_attribute('innerText')
                    tries = 20
                except Exception as e:
                    # TODO: Uncomment this to see if code error or scrapper error
                    #print(e)
                    tries += 1
                    time.sleep(0.250)
            else:
                break
    finally:
        driver.close()

    return translation

    """
    # Intersting options: https://github.com/ultrafunkamsterdam/undetected-chromedriver/issues/1726
    # Also this: https://stackoverflow.com/questions/19211006/how-to-enable-cookies-in-chromedriver-with-webdriver
    # What about this: options.AddUserProfilePreference("profile.cookie_controls_mode", 0);
    """


from yta_general_utils.programming.path import get_project_abspath
from dotenv import load_dotenv

import os


# TODO: Create a Environment class to wrap maybe (?)
def load_current_project_dotenv():
    """
    This methods looks for the '.env' file in the project that is
    currently being executed (the code, not the library) and loads
    it.

    This means that any project in which you are importing this 
    library, the '.env' file in its main folder will be loaded.
    """
    dotenv_path = os.path.join(get_project_abspath(), '.env')
    load_dotenv(dotenv_path)

def get_current_project_env(variable: str):
    """
    Loads the current project environment '.env' configuration file
    and returns, if available, the 'variable' variable value.

    This method does a 'load_dotenv' method call within the current
    project absolute path any time you call it, so it ensures 
    loading the values correctly if available.
    
    You can use directly this method instead of 'os.getenv()'
    without using any 'load_dotenv()' method before because it is 
    implicitely called within the current project.
    """
    load_current_project_dotenv()

    return os.getenv(variable)



"""
THIS IS VERY IMPORTANT TO UNDERSTAND THE CODE:

- First of all, the distance. Imagine the map
once it's been plotted. If you go from the origin
to the end, there is a distance you are virtually
drawing. That is the global distance. The distance
from the origin, normalized, is d=1.0. The origin
is d=0.0 and the end is d=1.0.

- As you are using pairs of coordinates, there is
also a local distance in between each pair of
coordinates. Imagine, for a second, that each pair
of coordinates is a map by itself. The first 
coordinate is the d=0.0 and the second coordinate
is the d=1.0 (in local distance terms).

- If you have, for example, 5 pairs of coordinates,
as the total global distance is d=1.0, each pair of
coordinates will represent a 1/5 of that total
global distance, so each pair of coordinates 
local distance is a 1/5 = 0.2 of the total global
distance. Knowing that, in order, the pairs of
coordinates represent the global distance as
follows: [0.0, 0.2], (0.2, 0.4], (0.4, 0.6],
(0.6, 0.8] and (0.8, 1.0] (for the same example
with 5 pairs of coordinates).

- Now that you have clear the previous steps, if
you think about a global distance of d=0.3, that
will be in the second pair of coordinates (the
one that represents (0.2, 0.4] range. But that
d=0.3 is in terms of global distance, so we need
to adapt it to that pair of coordinates local
distance. As we skipped 1 (the first) pair of
coordinates, we need to substract its distance
representation from our global distance, so
d=0.3 - 0.2 => d=0.1. Now, as d=0.1 is a global
distance value, we need to turn it into a local
distance value. As each pair of coordinates
represents a 0.2 of the global distance, we do
this: d=0.1 / 0.2 => d=0.5 and we obtain a local
distance of d=0.5. That is the local distance
within the second pair of coordinates (in this
example) we need to look for.

- Once we know the local distance we need to 
look for, as we now the pair of coordinates X
value of each of those coordinates, we can 
calculate the corresponding X value for what
that local distance fits.

As you can see, we go from a global distance to
a local X value to obtain the corresponding Y
of the affected pair of coordinates. This class
has been created with the purpose of following
a movement, so the distance we are talking about
is actually the amount of that movement we have
done previously so we can obtain the next 
position to which we need to move to follow the
movement that this map describes.
"""
from yta_general_utils.coordinate.coordinate import Coordinate, validate_coordinate_position
from yta_general_utils.math.map.pair_of_coordinates import PairOfCoordinates
from yta_general_utils.math.progression import Progression
from typing import Union

import matplotlib.pyplot as plt


# TODO: I think this is not actually a Map so the
# class name should change as it is more a Movement
# graphic and the Graphic class is an Animation
# Graphic or something like that
class Map:
    """
    A class to represent an scenario in which we will
    position Coordinates. The order of the coordinates
    will determine the direction of the movement, and
    they can be disordered (the next coordinate can be
    in a lower X and Y position). Also, there can be
    different coordinates in the same position.

    This is useful to simulate the movement within a
    scene.
    """
    
    coordinates: list[Coordinate] = None

    @property
    def pairs_of_coordinates(self) -> list[PairOfCoordinates]:
        """
        Get pairs of coordinates ordered by the moment in which
        they were added to the map. There is, at least, one pair
        of coordinates that will be, if no more coordinates
        added, the first and the last one.
        """
        return [
            PairOfCoordinates(self.coordinates[index], self.coordinates[index + 1])
            for index, _ in enumerate(self.coordinates[1:])
        ]

    @property
    def min_x(self):
        return min(self.coordinates, key = lambda coordinate: coordinate.x).x

    @property
    def max_x(self):
        return max(self.coordinates, key = lambda coordinate: coordinate.x).x
    
    @property
    def min_y(self):
        return min(self.coordinates, key = lambda coordinate: coordinate.y).y

    @property
    def max_y(self):
        return max(self.coordinates, key = lambda coordinate: coordinate.y).y

    def __init__(self):
        self.coordinates = []

    def add_coordinate(self, position: Union[tuple[float, float], list[float]]):
        """
        Add a new coordinate to the map if valid. The
        coordinate added can be in the same position as
        another coordinate.
        
        This method returns the instance so you can chain
        more than one 'add_coordinate' method call.
        """
        # TODO: Maybe accept Coordinate instance also
        # validating its position
        validate_coordinate_position(position)

        self.coordinates.append(Coordinate(position))

        return self
    
    def get_n_values(self, n: int):
        """
        Return a list of 'n' (x, y) values of the map
        from the start to the end of it.
        """
        return [
            self.get_xy_from_normalized_global_d(d)
            for d in Progression(0, 1, n).values
        ]
    
    def get_xy_from_normalized_global_d(self, d: float):
        """
        Return a not normalized tuple (x, y) representing
        the position for the provided global distance 'd',
        that is the global distance in the whole graphic X
        axis. The 'x' value of the first coordinate of the
        would be d=0.0 and the 'x' value of the last 
        coordinate, d=1.0.
        """
        # I'm using only this method as public method
        # because the 'd' global value is the only one that
        # lets me calculate the 'x' without a rounding
        # error.
        pair_of_coordinates = self._get_pair_of_coordinates_from_normalized_global_d(d)

        return pair_of_coordinates.get_xy_from_normalized_local_d(self._get_normalized_local_d_from_normalized_global_d(d))

    # Helpers below
    # TODO: Many helpers. Necessary (?)
    def _get_pair_of_coordinates_d(self):
        """
        Get the global distance 'd' that each pair of
        coordinates occupy.
        """
        return 1 / len(self.pairs_of_coordinates)
    
    def _get_pair_of_coordinates_index_from_normalized_global_d(self, d: float):
        """
        The 'd' is a global normalized distance value.
        """
        return int(d // self._get_pair_of_coordinates_d()) if d != 1.0 else len(self.pairs_of_coordinates) - 1
    
    def _get_pair_of_coordinates_from_normalized_global_d(self, d: float):
        """
        Look for the pair of coordinates corresponding to
        the given global distance 'd' value.
        """
        pair_of_coordinates_index = self._get_pair_of_coordinates_index_from_normalized_global_d(d)
        return self.pairs_of_coordinates[pair_of_coordinates_index]

    def _get_normalized_local_d_from_normalized_global_d(self, d: float):
        """
        Transform the global 'd' to a local 'd' that would
        be in the corresponding pair of coordinates.
        """
        pair_of_coordinates_index = self._get_pair_of_coordinates_index_from_normalized_global_d(d)
        pair_of_coordinates_d = self._get_pair_of_coordinates_d()

        if d != 1.0:
            d = (
                d % (pair_of_coordinates_index * pair_of_coordinates_d) / pair_of_coordinates_d
                if pair_of_coordinates_index > 0 else d / pair_of_coordinates_d
            )

        return d
    # Helpers above
    
    def plot(self):
        # Limit and draw axis
        # plt.xlim(self.min_x, self.max_x)
        # plt.ylim(self.min_y, self.max_y)
        plt.xlim(0, 1920)
        plt.ylim(0, 1080)
        plt.axhline(0, color = 'black', linewidth = 1)
        plt.axvline(0, color = 'black', linewidth = 1)

        plt.grid(True)

        # Draw nodes
        x_vals = [coordinate.x for coordinate in self.coordinates]
        y_vals = [coordinate.y for coordinate in self.coordinates]
        plt.scatter(x_vals, y_vals, color = 'white', edgecolors = 'black', s = 100)

        # Draw points between nodes
        xs = []
        ys = []
        for pair_of_coordinates in self.pairs_of_coordinates:
            positions = pair_of_coordinates.get_n_xy_values_to_plot(100)
            t_xs, t_ys = zip(*positions)
            xs += t_xs
            ys += t_ys
       
        plt.scatter(xs, ys, color = 'black', s = 1)

        # Ploting is from bottom-left to top-right while moviepy
        # scenario is from top-left to bottom-right. We need to
        # invert the Y axis to see the graphic as the movement
        plt.gca().invert_yaxis()
        
        plt.title('')
        plt.xlabel('X axis')
        plt.ylabel('Y axis')
        
        plt.show()


from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.math.rate_functions.rate_function import RateFunction
from typing import Union


class RateFunctionArgument:
    """
    An argument that can act as a rate function
    modifying the calculations for the animations
    we need. Anything that is able to recalculate
    an 'n' normalized value according and return
    it is accepted here.

    Check the accepted types in the '__init__'
    method.
    """

    def __init__(self, argument: Union['RateFunction', 'RateFunctionGraphic']):
        if not PythonValidator.is_instance(argument, ['RateFunction', 'RateFunctionGraphic']):
            raise Exception('The provided "argument" parameter is not an instance of RateFunction nor RateFunctionGraphic class.')
        
        self.argument = argument
    
    def get_n_value(self, n: float):
        """
        Get the corresponding value for the given 'n' normalized
        value that must be between 0 and 1.
        """
        return self.argument.get_n_value(n)
    
    @staticmethod
    def default():
        """
        Return a default instance of this class, perfect to
        be used as a parameter default value.
        """
        return RateFunctionArgument(RateFunction.LINEAR)


from yta_general_utils.math.graphic.node import Node
from yta_general_utils.math.normalizable_value import NormalizableValue
from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from yta_general_utils.math.rate_functions.rate_function import RateFunction
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.math.progression import Progression


# This class is very similar to PairOfCoordinates,
# maybe we can do something to avoid the duplicated
# code
class PairOfNodes:
    """
    Class to represent a pair of consecutive Nodes
    within a Graphic, that are connected and able
    to calculate any 'd' value between them. The
    left node must be positioned in a lower X than 
    the right one to be consecutive and valid.

    This pair of nodes will be represented by the 0
    to 1 X and Y axis values locally so they can be
    turned into the general Graphic value applying
    the general Graphic limits.
    """
    
    left_node: Node = None
    right_node: Node = None
    rate_function: RateFunctionArgument = None
    """
    The rate function to join the left node with the
    right node.
    """

    @property
    def max_x(self):
        """
        The maximum X value, which is the right node X value.
        """
        return self.right_node.x
    
    @property
    def min_x(self):
        """
        The minimum X value, which is the left node X value.
        """
        return self.left_node.x
    
    @property
    def max_y(self):
        """
        The maximum Y value, which can be the left node or
        right node Y value.
        """
        return max([self.left_node.y, self.right_node.y])
    
    @property
    def min_y(self):
        """
        The minimum Y value, which can be the left node or
        right node Y value.
        """
        return min([self.left_node.y, self.right_node.y])
    
    @property
    def is_descendant(self):
        """
        Check if the Y value of the left node is greater
        than the Y value of the right node. If so, the
        pair of nodes is descendant (in Y value).
        """
        return self.left_node.y > self.right_node.y

    def __init__(
        self,
        left_node: Node,
        right_node: Node,
        rate_function: RateFunctionArgument = RateFunctionArgument(RateFunction.EASE_IN_OUT_SINE)
    ):
        if left_node.x > right_node.x:
            raise Exception('The left_node "x" value must be lower than the right_node "x" value.')
        
        self.left_node = left_node
        self.right_node = right_node
        self.rate_function = rate_function

    def get_n_xy_values_to_plot(
        self,
        n: int = 100,
        do_normalize: bool = False
    ) -> list[tuple[NormalizableValue, NormalizableValue]]:
        """
        Return 'n' (x, y) values to be plotted. Each of those
        X and Y values are normalized only if 'do_normalize'
        flag parameter is set as True.
        """
        if not NumberValidator.is_positive_number(n):
            raise Exception('The provided "n" parameter is not a positive number.')
        
        n = int(n)

        xs = [
            NormalizableValue(x, (self.min_x, self.max_x))
            for x in Progression(self.min_x, self.max_x, 100, RateFunctionArgument.default()).values
        ]
        ys = [
            self.get_y_from_not_normalized_x(x.value)
            for x in xs
        ]

        if do_normalize:
            xs = [x.normalized for x in xs]
            ys = [y.normalized for y in ys]
        else:
            xs = [x.value for x in xs]
            ys = [y.value for y in ys]

        return list(zip(xs, ys))

    def get_y_from_not_normalized_x(
        self,
        x: float
    ) -> NormalizableValue:
        """
        The X parameter must be a value between the left
        node X and the right node X.
        """
        return self._get_y_from_x(x, is_x_normalized = False)
    
    def get_y_from_normalized_x(
        self,
        x: float
    ) -> NormalizableValue:
        """
        The X parameter must be a value between 0 and 1,
        being 0 the left node X and 1 the right node X.
        """
        return self._get_y_from_x(x, is_x_normalized = True)
    
    def _get_y_from_x(
        self,
        x: float,
        is_x_normalized: bool = False
    ) -> NormalizableValue:
        """
        Get the Y value for the given X, depending on if
        the X value is normalized or not, flagged with the
        'is_x_normalized' parameter.
        
        This method is for internal use only.
        """
        lower_limit = self.min_x if not is_x_normalized else 0
        upper_limit = self.max_x if not is_x_normalized else 1

        if not NumberValidator.is_number_between(x, lower_limit, upper_limit):
            raise Exception(f'The "x" parameter must be between [{lower_limit}, {upper_limit}].')
        
        value = NormalizableValue(x, (self.min_x, self.max_x), value_is_normalized = is_x_normalized)
        value = NormalizableValue(self.rate_function.get_n_value(value.normalized), (self.min_y, self.max_y), value_is_normalized = True)
        value = NormalizableValue(1 - value.normalized, (self.min_y, self.max_y), value_is_normalized = True) if self.is_descendant else value

        return value


from yta_general_utils.math.graphic.graphic import Graphic
from yta_general_utils.math.value_normalizer import ValueNormalizer
from yta_general_utils.programming.parameter_validator import PythonValidator


class RateFunctionGraphic:
    """
    A Graphic instance that uses itself as a rate
    function and returns the corresponding value
    for the given 'n'.

    TODO: Maybe force that the provided graphic 
    must end in the maximum Y value so, when
    normalized, it ends in value 1.
    """
    def __init__(self, graphic: Graphic):
        if not PythonValidator.is_instance(graphic, Graphic):
            raise Exception('The "graphic" parameter provided is not a Graphic class instance.')

        self.graphic = graphic
        
    def get_n_value(self, n: float):
        """
        Get the corresponding value for the given 'n' normalized
        value that must be between 0 and 1.
        """
        return ValueNormalizer(self.graphic.min_y, self.graphic.max_y).normalize(self.graphic.get_xy_from_normalized_d(n)[1])


from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.math import Math


# TODO: Maybe create a Value class to be able to
# handle if it has been normalized or not and to
# also handle its range? It could be interesting
# for Progression class, but it is too much I
# think...

# TODO: What about inverted limits and ranges (?)
class ValueNormalizer:
    """
    Class to simplify the process of obtaining normalized
    values between a range and also converting them into
    a value within a new range.

    The range of normalized values will be always [0, 1].
    """
    def __init__(self, lower_limit: float, upper_limit: float):
        if lower_limit > upper_limit:
            raise Exception('Review the limits. The lower limit cannot be equal or greater to the upper limit.')

        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def _validate_outlier_value(self, value: float):
        """
        Validate if the provided 'value' is within the limits
        or if it is an outlier, raising an Exception it an
        outlier value is detected.
        """
        if not NumberValidator.is_number_between(value, self.lower_limit, self.upper_limit):
            raise Exception(f'The provided "value" is out of the limits [{str(self.lower_limit)}, {str(self.upper_limit)}] and this outliers are not accepted due to "do_accept_outliers" parameter False value.')
        
    def _validate_normalized_outlier_value(self, value: float):
        """
        Validate if the provided 'value' is within the 
        normalized limits or if it is an outlier, raising
        an Exception it an outlier value is detected.
        """
        if not NumberValidator.is_number_between(value, 0, 1):
            raise Exception(f'The provided "value" is out of the limits [0, 1] and this outliers are not accepted due to "do_accept_outliers" parameter False value.')

    def normalize(self, value: float, do_accept_outliers: bool = True):
        """
        Obtain the provided 'value' as a normalized value,
        that will be between 0 and 1 according to this
        instance lower and upper limit. The provided
        'value' must be a not normalized value.

        If 'do_accept_outliers' is True, giving a value
        out of the range will also return its corresponding
        normalized value.
        
        A value equal than the lower limit would be 0, and
        a value equal to the upper limit, 1.

        15 is 0.5 in [0, 1] range for [10, 20] limits.
        """
        if not NumberValidator.is_number(value):
            raise Exception('The provided "value" is not a number.')
        
        if not do_accept_outliers:
            self._validate_outlier_value(value)

        if self.lower_limit == self.upper_limit:
            return 1.0

        return Math.normalize(value, self.lower_limit, self.upper_limit)
    
    def denormalize(self, value: float, do_accept_outliers: bool = True):
        """
        Obtain the provided normalized 'value' as a value
        that will be between this instance lower and upper
        limit.  The provided 'value' must be a normalized
        value.

        If 'do_accept_outliers' is True, giving a value
        out of the normalized range will also return its
        corresponding value according to the original range
        (that will be out of it, obviously).
        
        A value equal than 0 would be the lower limit, and
        a value equal to 1, the upper limit.

        0.5 is 15 is [10, 20] range for [0, 1] normalized 
        limits.
        """
        if not NumberValidator.is_number(value):
            raise Exception('The provided "value" is not a number.')
        
        if not do_accept_outliers:
            self._validate_normalized_outlier_value(value)

        if self.lower_limit == self.upper_limit:
            return self.lower_limit

        return Math.denormalize(value, self.lower_limit, self.upper_limit)
    
    def to_range_value(self, value: float, new_range_lower_limit: float, new_range_upper_limit: float, do_accept_outliers: bool = True):
        """
        Obtain the provided 'value' as its corresponding 
        value within a new range in between the provided
        'new_range_lower_limit' and 'new_range_upper_limit'.
        The provided 'value' must be a not normalized value.

        If 'do_accept_outliers' is True, giving a value
        out of the range will also return its corresponding
        value in the new range.

        If lower limit is 10 and upper limit is 20, and a 
        value of 15 is provided, the normalized value would
        be 0.5, but as we want in a new range, we transform
        it. Asking for a new range lower limit of 0 and a 
        new range upper limit of 100, the same 15 value will
        return 50 in this method.

        15 is 0.5 in [0, 1] range for [10, 20] limits, but
        50 in the new [0, 100] range.
        """
        if new_range_lower_limit == new_range_upper_limit or new_range_lower_limit > new_range_upper_limit:
            raise Exception('Review the limits. The new range lower limit cannot be equal or greater to the new range upper limit.')
        
        if not do_accept_outliers:
            self._validate_outlier_value(value)
        
        return Math.denormalize(self.normalize(value), new_range_lower_limit, new_range_upper_limit)